from __future__ import annotations
from typing import TypedDict, List, Any, TypeAlias, Literal
from typing_extensions import Required, NotRequired


class ChatMessage(TypedDict, total=True):
    role: Required[Literal["system", "user", "assistant"]]
    content: Required[str]


class ComparisonCompletion(TypedDict, total=True):
    text: Required[str]
    model: Required[str]


class CompletionComparisonFilterInput(TypedDict, total=True):
    """
    Filter for completion preference feedbacks.

    Args:
        metric: Feedback key logged against.
    """

    metric: Required[str]


class CompletionFeedbackFilterInput(TypedDict, total=False):
    """
    Filter for completion metric feedbacks.

    Args:
        metric: Feedback key logged against.
        gt: >
        gte: >=
        eq: ==
        neq: !=
        lt: <
        lte: <=
        user: Feedbacks logged by `user` id.
    """

    metric: Required[str]
    gt: NotRequired[float]
    gte: NotRequired[float]
    eq: NotRequired[float]
    neq: NotRequired[float]
    lt: NotRequired[float]
    lte: NotRequired[float]
    reasons: NotRequired[List[str]]
    user: NotRequired[Any]


class CompletionLabelFilter(TypedDict, total=False):
    """
    Filter for completion labels.
    """

    key: Required[str]
    value: NotRequired[List[str]]


CompletionSource: TypeAlias = Literal["LIVE", "OFFLINE", "AUTOMATION", "DATASET"]


class CursorPageInput(TypedDict, total=False):
    """
    Paging config.

    Args:
        first: Retrieve first n items starting from the `after` cursor.
        last: Retrieve last n items starting from the `after` cursor, limited to `before` cursor.
        after: Start cursor.
        before: End cursor.
    """

    first: NotRequired[int]
    after: NotRequired[str]
    before: NotRequired[str]
    last: NotRequired[int]


class InteractionFeedbackDict(TypedDict):
    """
    Interaction feedback.

    Args:
        feedback_key: Feedback key to register feedback against.
        value: Metric feedback value.
        details: Optional feedback text details.
    """

    feedback_key: Required[str]
    value: Required[int | float | bool]
    details: NotRequired[str]


class JudgeExampleInput(TypedDict, total=False):
    """
    Example to guide an AI judge's reasoning when evaluating a completion (few-shot prompting).

    Args:
        input: Ordered list of chat messages (role/content) that form the conversation context.
        output: Assistant completion to be evaluated.
        passes: Boolean indicating whether the *output* satisfies the criteria.
        reasoning: Optional free-text with the rationale behind the decision.
    """

    input: Required[List[ChatMessage]]
    output: Required[str]
    passes: Required[bool]
    reasoning: Required[str]


class ListCompletionsFilterInput(TypedDict, total=False):
    """
    Filter for listing interactions.

    Args:
        models: Model keys.
        timerange: A timerange in timestamp format.
        user_id: User ID that created interaction.
        feedbacks: TypedDict for metric feedback filtering.
        comparisons: TypedDict for preference feedback filtering.
        labels: TypedDict for completion labels filtering.
        completion_id: Completion id.
        source: Interaction source filter.
    """

    models: NotRequired[List[str]]
    timerange: NotRequired["TimeRange"]
    session_id: NotRequired[Any]
    user_id: NotRequired[Any]
    feedbacks: NotRequired[List["CompletionFeedbackFilterInput"]]
    comparisons: NotRequired[List["CompletionComparisonFilterInput"]]
    labels: NotRequired[List["CompletionLabelFilter"]]
    prompt_hash: NotRequired[str]
    completion_id: NotRequired[Any]
    tags: NotRequired[List[str]]
    source: NotRequired[List[CompletionSource]]


class ModelComputeConfigInput(TypedDict, total=False):
    tp: NotRequired[int]
    kv_cache_len: NotRequired[int]
    max_seq_len: NotRequired[int]


class ModelFilter(TypedDict, total=False):
    in_storage: NotRequired[bool]
    available: NotRequired[bool]
    trainable: NotRequired[bool]
    kind: NotRequired[List[Literal["Embedding", "Generation"]]]
    view_all: NotRequired[bool]
    online: NotRequired[List[Literal["ONLINE", "OFFLINE", "PENDING", "ERROR"]]]


class ModelPlacementInput(TypedDict, total=False):
    compute_pools: Required[List[str]]
    max_ttft_ms: NotRequired[int]


class TimeRange(TypedDict, total=False):
    """
    A timerange filter, in Unix timestamp format (ms).

    Args:
        from_: The start timestamp.
        to: The end timestamp.
    """

    from_: Required[int | str]
    to: Required[int | str]


class Order(TypedDict, total=False):
    """
    Ordering of interaction list results.

    Args:
        field: On what field to order by.
        order: Ascending or descending; alphabetical for string fields.
    """

    field: Required[str]
    order: Required[Literal["ASC", "DESC"]]



from __future__ import annotations
from typing import TYPE_CHECKING, Literal, Any

from adaptive_sdk.graphql_client import (
    CursorPageInput,
    JobInput,
    ListJobsJobs,
    ListJobsFilterInput,
    JobData,
    JobKind,
)


from .base_resource import SyncAPIResource, AsyncAPIResource, UseCaseResource

if TYPE_CHECKING:
    from adaptive_sdk.client import Adaptive, AsyncAdaptive


class Jobs(SyncAPIResource, UseCaseResource):  # type: ignore[misc]
    """
    Resource to interact with jobs.
    """

    def __init__(self, client: Adaptive) -> None:
        SyncAPIResource.__init__(self, client)
        UseCaseResource.__init__(self, client)

    def get(self, job_id: str) -> JobData | None:
        return self._gql_client.describe_job(id=job_id).job

    def list(
        self,
        first: int | None = 100,
        last: int | None = None,
        after: str | None = None,
        before: str | None = None,
        kind: (
            list[
                Literal[
                    "TRAINING",
                    "EVALUATION",
                    "DATASET_GENERATION",
                    "MODEL_CONVERSION",
                    "CUSTOM",
                ]
            ]
            | None
        ) = None,
        use_case: str | None = None,
    ) -> ListJobsJobs:
        use_case = self.optional_use_case_key(use_case)
        page = CursorPageInput(first=first, last=last, after=after, before=before)
        validated_filter = ListJobsFilterInput(
            useCase=use_case,
            kind=[JobKind(k) for k in kind] if kind else None,
        )
        return self._gql_client.list_jobs(page=page, filter=validated_filter, order=[]).jobs

    def run(
        self,
        recipe_key: str,
        args: dict[str, Any],
        num_gpus: int,
        use_case: str | None = None,
        compute_pool: str | None = None,
    ) -> JobData:
        return self._gql_client.create_job(
            input=JobInput(
                recipe=recipe_key,
                args=args,
                useCase=self.use_case_key(use_case),
                computePool=compute_pool,
                numGpus=num_gpus,
            )
        ).create_job

    def cancel(self, job_id: str) -> JobData:
        return self._gql_client.cancel_job(job_id=job_id).cancel_job


class AsyncJobs(AsyncAPIResource, UseCaseResource):  # type: ignore[misc]
    """
    Async resource to interact with external rewards servers.
    """

    def __init__(self, client: AsyncAdaptive) -> None:
        AsyncAPIResource.__init__(self, client)
        UseCaseResource.__init__(self, client)

    async def get(self, job_id: str) -> JobData | None:
        return (await self._gql_client.describe_job(id=job_id)).job

    async def list(
        self,
        first: int | None = 100,
        last: int | None = None,
        after: str | None = None,
        before: str | None = None,
        kind: (
            list[
                Literal[
                    "TRAINING",
                    "EVALUATION",
                    "DATASET_GENERATION",
                    "MODEL_CONVERSION",
                    "CUSTOM",
                ]
            ]
            | None
        ) = None,
        use_case: str | None = None,
    ) -> ListJobsJobs:
        page = CursorPageInput(first=first, last=last, after=after, before=before)
        validated_filter = ListJobsFilterInput(
            useCase=self.use_case_key(use_case),
            kind=[JobKind[k] for k in kind] if kind else None,
        )
        return (await self._gql_client.list_jobs(page=page, filter=validated_filter)).jobs

    async def run(
        self,
        recipe_key: str,
        args: dict[str, Any],
        num_gpus: int,
        use_case: str | None = None,
        compute_pool: str | None = None,
    ) -> JobData:
        return (
            await self._gql_client.create_job(
                input=JobInput(
                    recipe=recipe_key,
                    args=args,
                    useCase=self.use_case_key(use_case),
                    computePool=compute_pool,
                    numGpus=num_gpus,
                )
            )
        ).create_job

    async def cancel(self, job_id: str) -> JobData:
        return (await self._gql_client.cancel_job(job_id=job_id)).cancel_job



# type: ignore

from __future__ import annotations
from typing import TYPE_CHECKING

from .base_resource import SyncAPIResource, AsyncAPIResource, UseCaseResource

if TYPE_CHECKING:
    from adaptive_sdk.client import Adaptive, AsyncAdaptive


class ComputePools(SyncAPIResource, UseCaseResource):
    """
    Resource to interact with compute pools.
    """

    def __init__(self, client: Adaptive) -> None:
        SyncAPIResource.__init__(self, client)
        UseCaseResource.__init__(self, client)

    def list(self):
        return self._gql_client.list_compute_pools()


class AsyncComputePools(AsyncAPIResource, UseCaseResource):
    """
    Resource to interact with compute pools.
    """

    def __init__(self, client: AsyncAdaptive) -> None:
        AsyncAPIResource.__init__(self, client)
        UseCaseResource.__init__(self, client)

    async def list(self):
        return await self._gql_client.list_compute_pools()



from pydantic import Field
from .base_model import BaseModel
from .fragments import GraderData

class LockGrader(BaseModel):
    """@public"""
    lock_grader: 'LockGraderLockGrader' = Field(alias='lockGrader')

class LockGraderLockGrader(GraderData):
    """@public"""
    pass
LockGrader.model_rebuild()


from .base_model import BaseModel
from .fragments import GraderData

class GetGrader(BaseModel):
    """@public"""
    grader: 'GetGraderGrader'

class GetGraderGrader(GraderData):
    """@public"""
    pass
GetGrader.model_rebuild()


from typing import List
from pydantic import Field
from .base_model import BaseModel
from .fragments import RemoteEnvData

class ListRemoteEnvs(BaseModel):
    """@public"""
    remote_envs: List['ListRemoteEnvsRemoteEnvs'] = Field(alias='remoteEnvs')

class ListRemoteEnvsRemoteEnvs(RemoteEnvData):
    """@public"""
    pass
ListRemoteEnvs.model_rebuild()


import httpx
from typing import Callable
from pydantic import BaseModel
from .graphql_client.exceptions import GraphQLClientGraphQLMultiError


class EntityNotFoundDescriptor(BaseModel):
    entity: str
    error_message: str
    code_suggestions: list[str]

    def __post_init__(self):
        if self.error_message.count("{}") != len(self.code_suggestions):
            raise ValueError("# of code suggestions must match # of placeholders in error message template")


class AdaptiveEntityNotFoundError(Exception):
    def __init__(self, original_error: str):
        self.entity_descriptors = [
            EntityNotFoundDescriptor(
                entity="use_case",
                error_message="Use case key does not exist. You can create it with `{}`",
                code_suggestions=["client.use_cases.create"],
            ),
            EntityNotFoundDescriptor(
                entity="metric",
                error_message="Feedback key does not exist, or is not linked to the target use case. You can create it with `{}` or link it with `{}`",
                code_suggestions=[
                    "client.feedback.register_key",
                    "client.feedback.link",
                ],
            ),
            EntityNotFoundDescriptor(
                entity="model",
                error_message="Model key does not exist, or is not attached to the target use case. You can attach it with `{}`",
                code_suggestions=["client.models.attach"],
            ),
            EntityNotFoundDescriptor(entity="user", error_message="User does not exist.", code_suggestions=[]),
        ]

        new_error_message = None
        for descriptor in self.entity_descriptors:
            if descriptor.entity in original_error.lower():
                new_error_message = descriptor.error_message.format(*descriptor.code_suggestions)
                break

        super().__init__(new_error_message or original_error)


def adaptive_error_message_triage(error_message: str) -> Callable | None:
    if "EntityNotFound:" in error_message:
        return AdaptiveEntityNotFoundError
    else:
        return None


def graphql_multi_error_handler(fn: Callable) -> Callable:
    def wrapper(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except GraphQLClientGraphQLMultiError as e:
            error_message = str(e.errors[0])
            exception_type = adaptive_error_message_triage(error_message)
            if exception_type is not None:
                raise exception_type(error_message)
            else:
                raise

    return wrapper


def rest_error_handler(response: httpx.Response):
    if response.is_client_error:
        try:
            response_json = response.json()
            error_message = response_json.get("error", f"Unknown error: {response.text}")
            exception_type = adaptive_error_message_triage(error_message)
            if exception_type is not None:
                raise exception_type(error_message)
            else:
                raise Exception(f"Client error {response.status_code}: {error_message}")
        except Exception:
            # Handle any exception when processing the response
            raise Exception(f"Client error {response.status_code}: {response.text}")
    else:
        response.raise_for_status()



from pydantic import Field
from .base_model import BaseModel
from .fragments import CustomRecipeData

class CreateCustomRecipe(BaseModel):
    """@public"""
    create_custom_recipe: 'CreateCustomRecipeCreateCustomRecipe' = Field(alias='createCustomRecipe')

class CreateCustomRecipeCreateCustomRecipe(CustomRecipeData):
    """@public"""
    pass
CreateCustomRecipe.model_rebuild()


from pydantic import Field
from .base_model import BaseModel
from .fragments import JudgeData

class UpdateJudge(BaseModel):
    """@public"""
    update_judge: 'UpdateJudgeUpdateJudge' = Field(alias='updateJudge')

class UpdateJudgeUpdateJudge(JudgeData):
    """@public"""
    pass
UpdateJudge.model_rebuild()


from pydantic import Field
from .base_model import BaseModel
from .fragments import UseCaseData

class ShareUseCase(BaseModel):
    """@public"""
    share_use_case: 'ShareUseCaseShareUseCase' = Field(alias='shareUseCase')

class ShareUseCaseShareUseCase(UseCaseData):
    """@public"""
    pass
ShareUseCase.model_rebuild()


from __future__ import annotations
from pathlib import Path
from typing import List, Literal, TYPE_CHECKING

from adaptive_sdk.graphql_client import (
    DatasetCreate,
    Upload,
    LoadDatasetCreateDataset,
    ListDatasetsDatasets,
    DatasetData,
)

from .base_resource import SyncAPIResource, AsyncAPIResource, UseCaseResource

if TYPE_CHECKING:
    from adaptive_sdk.client import Adaptive, AsyncAdaptive


class Datasets(SyncAPIResource, UseCaseResource):  # type: ignore[misc]
    """
    Resource to interact with file datasets.
    """

    def __init__(self, client: Adaptive) -> None:
        SyncAPIResource.__init__(self, client)
        UseCaseResource.__init__(self, client)

    def upload(
        self,
        file_path: str,
        dataset_key: str,
        name: str | None = None,
        use_case: str | None = None,
    ) -> LoadDatasetCreateDataset:
        """
         Upload a dataset from a file. File must be jsonl, where each line should match structure in example below.

        Args:
            file_path: Path to jsonl file.
            dataset_key: New dataset key.
            name: Optional name to render in UI; if `None`, defaults to same as `dataset_key`.

        Example:
        ```
        {"messages": [{"role": "system", "content": "<optional system prompt>"}, {"role": "user", "content": "<user content>"}, {"role": "assistant", "content": "<assistant answer>"}], "completion": "hey"}
        ```
        """
        input = DatasetCreate(
            useCase=self.use_case_key(use_case),
            name=name if name else dataset_key,
            key=dataset_key,
        )
        filename = Path(file_path).stem
        with open(file_path, "rb") as f:
            file_upload = Upload(filename=filename, content=f, content_type="application/jsonl")
            return self._gql_client.load_dataset(input=input, file=file_upload).create_dataset

    def list(
        self,
        use_case: str | None = None,
    ) -> List[ListDatasetsDatasets]:
        """
        List previously uploaded datasets.
        """
        return self._gql_client.list_datasets(self.use_case_key(use_case)).datasets

    def get(self, key: str, use_case: str | None = None) -> DatasetData | None:
        """
        Get details for dataset.

        Args:
            key: Dataset key.
        """
        return self._gql_client.describe_dataset(key, self.use_case_key(use_case)).dataset

    def delete(self, key: str, use_case: str | None = None) -> bool:
        """Delete dataset."""
        return self._gql_client.delete_dataset(id_or_key=key, use_case=self.use_case_key(use_case)).delete_dataset


class AsyncDatasets(AsyncAPIResource, UseCaseResource):  # type: ignore[misc]
    def __init__(self, client: AsyncAdaptive) -> None:
        AsyncAPIResource.__init__(self, client)
        UseCaseResource.__init__(self, client)

    async def upload(
        self,
        file_path: str,
        dataset_key: str,
        name: str | None = None,
        use_case: str | None = None,
    ) -> LoadDatasetCreateDataset:
        """
        Upload a dataset from a file. File must be jsonl, where each line should match structure in example below.

        Args:
            file_path: Path to jsonl file.
            dataset_key: New dataset key.
            name: Optional name to render in UI; if `None`, defaults to same as `dataset_key`.

        Example:
        ```
        {"messages": [{"role": "system", "content": "<optional system prompt>"}, {"role": "user", "content": "<user content>"}, {"role": "assistant", "content": "<assistant answer>"}], "completion": "hey"}
        ```
        """
        input = DatasetCreate(
            useCase=self.use_case_key(use_case),
            name=name if name else dataset_key,
            key=dataset_key,
        )
        filename = Path(file_path).stem
        with open(file_path, "rb") as f:
            file_upload = Upload(filename=filename, content=f, content_type="application/jsonl")
            upload_result = await self._gql_client.load_dataset(input=input, file=file_upload)
            return upload_result.create_dataset

    async def list(
        self,
        use_case: str | None = None,
    ) -> List[ListDatasetsDatasets]:
        """
        List previously uploaded datasets.
        """
        results = await self._gql_client.list_datasets(self.use_case_key(use_case))
        return results.datasets

    async def get(self, key: str, use_case: str | None = None) -> DatasetData | None:
        """
        Get details for dataset.

        Args:
            key: Dataset key.
        """
        result = await self._gql_client.describe_dataset(key, self.use_case_key(use_case))
        return result.dataset

    async def delete(self, key: str, use_case: str | None = None) -> bool:
        """Delete dataset."""
        return (
            await self._gql_client.delete_dataset(id_or_key=key, use_case=self.use_case_key(use_case))
        ).delete_dataset



from typing import Optional
from .base_model import BaseModel
from .fragments import MetricData

class DescribeMetric(BaseModel):
    """@public"""
    metric: Optional['DescribeMetricMetric']

class DescribeMetricMetric(MetricData):
    """@public"""
    pass
DescribeMetric.model_rebuild()


from typing import List
from .base_model import BaseModel
from .fragments import UserData

class ListUsers(BaseModel):
    """@public"""
    users: List['ListUsersUsers']

class ListUsersUsers(UserData):
    """@public"""
    pass
ListUsers.model_rebuild()


from typing import Any, List
from pydantic import Field
from .base_model import BaseModel

class ListTeams(BaseModel):
    """@public"""
    teams: List['ListTeamsTeams']

class ListTeamsTeams(BaseModel):
    """@public"""
    id: Any
    key: str
    name: str
    created_at: int = Field(alias='createdAt')
ListTeams.model_rebuild()


from __future__ import annotations
from uuid import UUID
from typing import overload

from adaptive_sdk import input_types
from .rest import rest_types


@overload
def convert_optional_UUID(input_param: str | UUID) -> UUID: ...


@overload
def convert_optional_UUID(input_param: str | UUID | None) -> UUID | None: ...


def convert_optional_UUID(input_param: str | UUID | None) -> UUID | None:
    if not input_param:
        return None
    if isinstance(input_param, UUID):
        return input_param
    try:
        return UUID(input_param)
    except ValueError as e:
        raise e


def validate_comparison_completion(
    completion: str | UUID | input_types.ComparisonCompletion,
) -> UUID | rest_types.CompletionIdOrText1:

    if isinstance(completion, dict):
        return rest_types.CompletionIdOrText1(
            text=completion["text"], model=completion["model"]
        )
    else:
        return convert_optional_UUID(completion)


def get_full_model_path(use_case_key: str, model_key: str | None):
    return f"{use_case_key}{'/' + model_key if model_key is not None else ''}"



from pydantic import Field
from .base_model import BaseModel
from .fragments import GraderData

class UpdateGrader(BaseModel):
    """@public"""
    update_grader: 'UpdateGraderUpdateGrader' = Field(alias='updateGrader')

class UpdateGraderUpdateGrader(GraderData):
    """@public"""
    pass
UpdateGrader.model_rebuild()


from .base_model import BaseModel
from .fragments import JudgeData

class GetJudge(BaseModel):
    """@public"""
    judge: 'GetJudgeJudge'

class GetJudgeJudge(JudgeData):
    """@public"""
    pass
GetJudge.model_rebuild()


from __future__ import annotations
from typing import TYPE_CHECKING, Sequence

from adaptive_sdk.graphql_client import (
    UserData,
    TeamMemberSet,
    UpdateUserSetTeamMember,
    TeamMemberRemove,
    UserCreate,
    UserCreateTeamWithRole,
)

from .base_resource import SyncAPIResource, AsyncAPIResource

if TYPE_CHECKING:
    from adaptive_sdk.client import Adaptive, AsyncAdaptive


class Users(SyncAPIResource):  # type: ignore[misc]
    """
    Resource to manage users and permissions.
    """

    def __init__(self, client: Adaptive) -> None:
        SyncAPIResource.__init__(self, client)

    def me(self) -> UserData | None:
        """
        Get details of current user.
        """
        return self._gql_client.me().me

    def list(self) -> Sequence[UserData]:
        """
        List all users registered to Adaptive deployment.
        """
        return self._gql_client.list_users().users

    def create(
        self, email: str, name: str, teams_with_role: Sequence[tuple[str, str]]
    ) -> UserData:
        """
        Create a user and with preset teams and role
        """
        return self._gql_client.create_user(
            input=UserCreate(
                email=email,
                name=name,
                teams=[
                    UserCreateTeamWithRole(team=team, role=role)
                    for (team, role) in teams_with_role
                ],
            )
        ).create_user

    def add_to_team(self, email: str, team: str, role: str) -> UpdateUserSetTeamMember:
        """
        Update team and role for user.

        Args:
            email: User email.
            team: Key of team to which user will be added to.
            role: Assigned role

        """
        input = TeamMemberSet(user=email, team=team, role=role)
        return self._gql_client.update_user(input).set_team_member

    def remove_from_team(self, email: str, team: str) -> UserData:
        """
        Remove user from team.

        Args:
            email: User email.
            team: Key of team to remove user from.
        """
        input = TeamMemberRemove(user=email, team=team)
        return self._gql_client.remove_team_member(input).remove_team_member

    def delete(self, email: str):
        self._gql_client.delete_user(email)


class AsyncUsers(AsyncAPIResource):  # type: ignore[misc]
    """
    Resource to manage users and permissions.
    """

    def __init__(self, client: AsyncAdaptive) -> None:
        AsyncAPIResource.__init__(self, client)

    async def me(self) -> UserData | None:
        """
        Get details of current user.
        """
        result = await self._gql_client.me()
        return result.me

    async def list(self) -> Sequence[UserData]:
        """
        List all users registered to Adaptive deployment.
        """
        result = await self._gql_client.list_users()
        return result.users

    async def create(
        self, email: str, name: str, teams_with_role: Sequence[tuple[str, str]]
    ) -> UserData:
        """
        Create a user and with preset teams and role
        """
        result = await self._gql_client.create_user(
            input=UserCreate(
                email=email,
                name=name,
                teams=[
                    UserCreateTeamWithRole(team=team, role=role)
                    for (team, role) in teams_with_role
                ],
            )
        )
        return result.create_user

    async def add_to_team(
        self, email: str, team: str, role: str
    ) -> UpdateUserSetTeamMember:
        """
        Update team and role for user.

        Args:
            email: User email.
            team: Key of team to which user will be added to.
            role: Assigned role

        """
        input = TeamMemberSet(user=email, team=team, role=role)
        result = await self._gql_client.update_user(input)
        return result.set_team_member

    async def remove_from_team(self, email: str, team: str) -> UserData:
        """
        Remove user from team.

        Args:
            email: User email.
            team: Key of team to remove user from.
        """
        input = TeamMemberRemove(user=email, team=team)
        return (await self._gql_client.remove_team_member(input)).remove_team_member

    async def delete(self, email: str):
        await self._gql_client.delete_user(email)



from __future__ import annotations

from typing import Sequence, TYPE_CHECKING, Any, Literal

from adaptive_sdk.graphql_client import (
    GraderData,
    GraderCreateInput,
    CustomConfigInput,
    GraderConfigInput,
    GraderTypeEnum,
    JudgeConfigInput,
    PrebuiltConfigInput,
    RemoteConfigInput,
    PrebuiltCriteriaKey,
    MetricGetOrCreate,
    JudgeExampleInput,
    JudgeExampleInputTurnEntry,
    TestRemoteEnvTestRemoteEnvRemoteEnvTestOffline,
    TestRemoteEnvTestRemoteEnvRemoteEnvTestOnline,
    RemoteEnvCreate,
)
from adaptive_sdk import input_types

from .base_resource import SyncAPIResource, AsyncAPIResource, UseCaseResource

if TYPE_CHECKING:
    from adaptive_sdk.client import Adaptive, AsyncAdaptive


class GraderCreator(SyncAPIResource, UseCaseResource):  # type: ignore[misc]
    """Helper class for creating different types of graders with a clean API."""

    def __init__(self, client: Adaptive) -> None:
        SyncAPIResource.__init__(self, client)
        UseCaseResource.__init__(self, client)

    def binary_judge(
        self,
        *,
        key: str,
        criteria: str,
        judge_model: str,
        feedback_key: str,
        name: str | None = None,
        examples: list[input_types.JudgeExampleInput] | None = None,
        use_case: str | None = None,
    ) -> GraderData:
        """
        Creates a new judge-based grader.

        Args:
            key: Unique key for the grader.
            criteria: Natural-language explanation of what should be evaluated as a binary pass/fail.
            judge_model: Model key of the judge model.
            feedback_key: Key for the feedback this grader writes to.
            examples: List of annotated examples for few-shot prompting.
            name: Human-readable grader name. If omitted, derived from key.
            use_case: Explicit use-case key. Falls back to client.default_use_case.
        """
        # Parse examples if provided
        parsed_examples = []
        if examples:
            for ex in examples:
                if not all(k in ex for k in ("input", "output", "passes")):
                    raise ValueError("Each example must contain 'input', 'output', and 'passes' keys")
                input_turns = [JudgeExampleInputTurnEntry(role=t["role"], content=t["content"]) for t in ex["input"]]
                parsed_examples.append(
                    JudgeExampleInput(
                        input=input_turns,
                        output=ex["output"],
                        reasoning=ex.get("reasoning"),
                        **{"pass": bool(ex["passes"])},
                    )
                )

        # Create judge config
        judge_config = JudgeConfigInput(
            model=judge_model,
            criteria=criteria,
            examples=parsed_examples,
        )

        # Create grader config
        grader_config = GraderConfigInput(judge=judge_config)

        input_obj = GraderCreateInput(
            name=name or key,
            key=key,
            graderType=GraderTypeEnum.JUDGE,
            graderConfig=grader_config,
            metric=MetricGetOrCreate(existing=feedback_key),
        )

        return self._client._gql_client.create_grader(
            use_case=self.use_case_key(use_case), input=input_obj
        ).create_grader

    def prebuilt_judge(
        self,
        *,
        key: str,
        type: Literal["FAITHFULNESS"],
        judge_model: str,
        name: str | None = None,
        use_case: str | None = None,
    ) -> GraderData:
        """
        Creates a new prebuilt grader.

        Args:
            key: Unique key for the grader.
            type: Type of prebuild AI Judge Grader.
            judge_model: Model key of the judge model.
            name: Human-readable grader name. If omitted, derived from key.
            use_case: Explicit use-case key. Falls back to client.default_use_case.
        """
        # Create prebuilt config
        assert type == "FAITHFULNESS", "Only faithfulness prebuilt grader is supported for now"
        prebuilt_config = PrebuiltConfigInput(
            key=PrebuiltCriteriaKey(type),
            model=judge_model,
        )

        # Create grader config
        grader_config = GraderConfigInput(prebuilt=prebuilt_config)

        input_obj = GraderCreateInput(
            name=name or key,
            key=key,
            graderType=GraderTypeEnum.PREBUILT,
            graderConfig=grader_config,
            metric=None,
        )

        return self._client._gql_client.create_grader(
            use_case=self.use_case_key(use_case), input=input_obj
        ).create_grader

    def external_endpoint(
        self,
        *,
        key: str,
        url: str,
        feedback_key: str,
        name: str | None = None,
        use_case: str | None = None,
    ) -> GraderData:
        """
        Creates a new external feedback endpoint grader.

        Args:
            key: Unique key for the grader.
            url: URL of the remote grading service.
            feedback_key: Key for the feedback this grader writes to.
            name: Human-readable grader name. If omitted, derived from key.
            use_case: Explicit use-case key. Falls back to client.default_use_case.
        """
        # Create remote config
        remote_config = RemoteConfigInput(url=url)

        # Create grader config
        grader_config = GraderConfigInput(remote=remote_config)

        input_obj = GraderCreateInput(
            name=name or key,
            key=key,
            graderType=GraderTypeEnum.REMOTE,
            graderConfig=grader_config,
            metric=MetricGetOrCreate(existing=feedback_key),
        )

        return self._client._gql_client.create_grader(
            use_case=self.use_case_key(use_case), input=input_obj
        ).create_grader

    def custom(
        self,
        *,
        key: str,
        feedback_key: str,
        description: str | None = None,
        name: str | None = None,
        use_case: str | None = None,
    ) -> GraderData:
        """
        Creates a new custom grader.

        Args:
            key: Unique key for the grader.
            feedback_key: Key for the feedback this grader writes to.
            description: Description of what grader does.
            name: Human-readable grader name. If omitted, derived from key.
            use_case: Explicit use-case key. Falls back to client.default_use_case.
        """
        grader_config = GraderConfigInput(custom=CustomConfigInput(description=description))

        input_obj = GraderCreateInput(
            name=name or key,
            key=key,
            graderType=GraderTypeEnum.CUSTOM,
            graderConfig=grader_config,
            metric=MetricGetOrCreate(existing=feedback_key),
        )

        return self._client._gql_client.create_grader(
            use_case=self.use_case_key(use_case), input=input_obj
        ).create_grader


class AsyncGraderCreator(AsyncAPIResource, UseCaseResource):  # type: ignore[misc]
    """Async helper class for creating different types of graders with a clean API."""

    def __init__(self, client: AsyncAdaptive) -> None:
        AsyncAPIResource.__init__(self, client)
        UseCaseResource.__init__(self, client)

    async def binary_judge(
        self,
        *,
        key: str,
        criteria: str,
        judge_model: str,
        feedback_key: str,
        name: str | None = None,
        examples: list[input_types.JudgeExampleInput] | None = None,
        use_case: str | None = None,
    ) -> GraderData:
        """
        Creates a new judge-based grader.

        Args:
            key: Unique key for the grader.
            criteria: Natural-language explanation of what should be evaluated as a binary pass/fail.
            judge_model: Model key of the judge model.
            feedback_key: Key for the feedback this grader writes to.
            examples: List of annotated examples for few-shot prompting.
            name: Human-readable grader name. If omitted, derived from key.
            use_case: Explicit use-case key. Falls back to client.default_use_case.
        """
        # Parse examples if provided
        parsed_examples = []
        if examples:
            for ex in examples:
                if not all(k in ex for k in ("input", "output", "passes")):
                    raise ValueError("Each example must contain 'input', 'output', and 'passes' keys")
                input_turns = [JudgeExampleInputTurnEntry(role=t["role"], content=t["content"]) for t in ex["input"]]
                parsed_examples.append(
                    JudgeExampleInput(
                        input=input_turns,
                        output=ex["output"],
                        reasoning=ex.get("reasoning"),
                        **{"pass": bool(ex["passes"])},
                    )
                )

        # Create judge config
        judge_config = JudgeConfigInput(
            model=judge_model,
            criteria=criteria,
            examples=parsed_examples,
        )

        # Create grader config
        grader_config = GraderConfigInput(judge=judge_config)

        input_obj = GraderCreateInput(
            name=name or key,
            key=key,
            graderType=GraderTypeEnum.JUDGE,
            graderConfig=grader_config,
            metric=MetricGetOrCreate(existing=feedback_key),
        )

        return (
            await self._client._gql_client.create_grader(use_case=self.use_case_key(use_case), input=input_obj)
        ).create_grader

    async def prebuilt_judge(
        self,
        *,
        key: str,
        type: Literal["FAITHFULNESS"],
        judge_model: str,
        name: str | None = None,
        use_case: str | None = None,
    ) -> GraderData:
        """
        Creates a new prebuilt grader.

        Args:
            key: Unique key for the grader.
            type: Type of prebuild AI Judge Grader.
            judge_model: Model key of the judge model.
            name: Human-readable grader name. If omitted, derived from key.
            use_case: Explicit use-case key. Falls back to client.default_use_case.
        """
        # Create prebuilt config
        assert type == "FAITHFULNESS", "Only faithfulness prebuilt grader is supported for now"
        prebuilt_config = PrebuiltConfigInput(
            key=PrebuiltCriteriaKey(type),
            model=judge_model,
        )

        # Create grader config
        grader_config = GraderConfigInput(prebuilt=prebuilt_config)

        input_obj = GraderCreateInput(
            name=name or key,
            key=key,
            graderType=GraderTypeEnum.PREBUILT,
            graderConfig=grader_config,
            metric=None,
        )

        return (
            await self._client._gql_client.create_grader(use_case=self.use_case_key(use_case), input=input_obj)
        ).create_grader

    async def external_endpoint(
        self,
        *,
        key: str,
        url: str,
        feedback_key: str,
        name: str | None = None,
        use_case: str | None = None,
    ) -> GraderData:
        """
        Creates a new external feedback endpoint grader.

        Args:
            key: Unique key for the grader.
            url: URL of the remote grading service.
            feedback_key: Existing key for the feedback this grader writes to.
            name: Human-readable grader name. If omitted, derived from key.
            use_case: Explicit use-case key. Falls back to client.default_use_case.
        """
        # Create remote config
        remote_config = RemoteConfigInput(url=url)

        # Create grader config
        grader_config = GraderConfigInput(remote=remote_config)

        input_obj = GraderCreateInput(
            name=name or key,
            key=key,
            graderType=GraderTypeEnum.REMOTE,
            graderConfig=grader_config,
            metric=MetricGetOrCreate(existing=feedback_key),
        )

        return (
            await self._client._gql_client.create_grader(use_case=self.use_case_key(use_case), input=input_obj)
        ).create_grader

    async def custom(
        self,
        *,
        key: str,
        feedback_key: str,
        description: str | None = None,
        name: str | None = None,
        use_case: str | None = None,
    ) -> GraderData:
        """
        Creates a new custom grader.

        Args:
            key: Unique key for the grader.
            feedback_key: Key for the feedback this grader writes to.
            description: Description of what grader does.
            name: Human-readable grader name. If omitted, derived from key.
            use_case: Explicit use-case key. Falls back to client.default_use_case.
        """
        grader_config = GraderConfigInput(custom=CustomConfigInput(description=description))

        input_obj = GraderCreateInput(
            name=name or key,
            key=key,
            graderType=GraderTypeEnum.CUSTOM,
            graderConfig=grader_config,
            metric=MetricGetOrCreate(existing=feedback_key),
        )

        return (
            await self._client._gql_client.create_grader(use_case=self.use_case_key(use_case), input=input_obj)
        ).create_grader


class Graders(SyncAPIResource, UseCaseResource):  # type: ignore[misc]
    """Resource to interact with grader definitions used to evaluate model completions."""

    def __init__(self, client: Adaptive) -> None:
        SyncAPIResource.__init__(self, client)
        UseCaseResource.__init__(self, client)

        # Nested creator for clean API
        self.create = GraderCreator(client)

    def delete(self, *, grader_key: str, use_case: str | None = None) -> bool:
        """Delete a grader. Returns True on success."""
        result = self._gql_client.delete_grader(use_case=self.use_case_key(use_case), id=grader_key).delete_grader
        return result.success

    def lock(self, *, grader_key: str, locked: bool, use_case: str | None = None) -> GraderData:
        """Lock or unlock a grader.

        Args:
            grader_key: ID or key of the grader.
            locked: Whether to lock (True) or unlock (False) the grader.
            use_case: Explicit use-case key. Falls back to client.default_use_case.
        """
        return self._gql_client.lock_grader(
            use_case=self.use_case_key(use_case), id=grader_key, locked=locked
        ).lock_grader

    def list(self, *, use_case: str | None = None) -> Sequence[GraderData]:
        """List all graders for the given use case."""
        return self._gql_client.list_graders(use_case=self.use_case_key(use_case)).graders

    def get(self, *, grader_key: str, use_case: str | None = None) -> GraderData | None:
        """Retrieve a specific grader by ID or key."""
        return self._gql_client.get_grader(id=grader_key, use_case=self.use_case_key(use_case)).grader

    def test_external_endpoint(
        self, url: str
    ) -> TestRemoteEnvTestRemoteEnvRemoteEnvTestOnline | TestRemoteEnvTestRemoteEnvRemoteEnvTestOffline:
        """Test external endpoint to check if it is reachable from Adaptive and returns a valid response."""
        return self._gql_client.test_remote_env(RemoteEnvCreate(url=url)).test_remote_env


class AsyncGraders(AsyncAPIResource, UseCaseResource):  # type: ignore[misc]
    """Asynchronous resource to interact with grader definitions used to evaluate model completions."""

    def __init__(self, client: AsyncAdaptive) -> None:
        AsyncAPIResource.__init__(self, client)
        UseCaseResource.__init__(self, client)

        # Nested creator for clean API
        self.create = AsyncGraderCreator(client)

    async def delete(self, *, grader_key: str, use_case: str | None = None) -> bool:
        """Delete a grader. Returns True on success."""
        result = (
            await self._gql_client.delete_grader(use_case=self.use_case_key(use_case), id=grader_key)
        ).delete_grader
        return result.success

    async def lock(self, *, grader_key: str, locked: bool, use_case: str | None = None) -> GraderData:
        """Lock or unlock a grader."""
        return (
            await self._gql_client.lock_grader(use_case=self.use_case_key(use_case), id=grader_key, locked=locked)
        ).lock_grader

    async def list(self, *, use_case: str | None = None) -> Sequence[GraderData]:
        """List all graders for the given use case."""
        return (await self._gql_client.list_graders(use_case=self.use_case_key(use_case))).graders

    async def get(self, *, grader_key: str, use_case: str | None = None) -> GraderData | None:
        """Retrieve a specific grader by ID or key."""
        return (await self._gql_client.get_grader(id=grader_key, use_case=self.use_case_key(use_case))).grader

    async def test_external_endpoint(
        self, url: str
    ) -> TestRemoteEnvTestRemoteEnvRemoteEnvTestOnline | TestRemoteEnvTestRemoteEnvRemoteEnvTestOffline:
        """Test external endpoint to check if it is reachable from Adaptive and returns a valid response."""
        return (await self._gql_client.test_remote_env(RemoteEnvCreate(url=url))).test_remote_env



from pydantic import Field
from .base_model import BaseModel
from .fragments import ModelData

class AddExternalModel(BaseModel):
    """@public"""
    add_external_model: 'AddExternalModelAddExternalModel' = Field(alias='addExternalModel')

class AddExternalModelAddExternalModel(ModelData):
    """@public"""
    pass
AddExternalModel.model_rebuild()


from __future__ import annotations
from typing import Literal, TYPE_CHECKING
from uuid import UUID
from typing_extensions import override
from adaptive_sdk.error_handling import rest_error_handler
from adaptive_sdk.rest import rest_types
from adaptive_sdk.utils import convert_optional_UUID, get_full_model_path

from .base_resource import SyncAPIResource, AsyncAPIResource, UseCaseResource

if TYPE_CHECKING:
    from adaptive_sdk.client import Adaptive, AsyncAdaptive

ROUTE = "/embeddings"


class Embeddings(SyncAPIResource, UseCaseResource):  # type: ignore[misc]
    """
    Resource to interact with embeddings.
    """

    def __init__(self, client: Adaptive) -> None:
        SyncAPIResource.__init__(self, client)
        UseCaseResource.__init__(self, client)

    def create(
        self,
        input: str,
        model: str | None = None,
        encoding_format: Literal["Float", "Base64"] = "Float",
        use_case: str | None = None,
        user: str | UUID | None = None,
    ) -> rest_types.EmbeddingsResponseList:
        """
        Creates embeddings inference request.

        Args:
            input: Input text to embed.
            model: Target model key for inference. If `None`, the requests will be routed to the use case's default model.
                Request will error if default model is not an embedding model.
            encoding_format: Encoding format of response.
            user: ID of user making the requests. If not `None`, will be logged as metadata for the request.
        """

        encoding_format_enum = rest_types.EmbeddingsEncodingFormat(encoding_format)
        emb_input = rest_types.GenerateEmbeddingsInput(
            input=input,
            model=get_full_model_path(self.use_case_key(use_case), model),
            encoding_format=encoding_format_enum,
            dimensions=None,
            user=convert_optional_UUID(user),
        )
        r = self._rest_client.post(
            ROUTE, data=emb_input.model_dump_json(exclude_none=True), headers={"Content-Type": "application/json"}  # type: ignore
        )
        rest_error_handler(r)
        return rest_types.EmbeddingsResponseList.model_validate(r.json())


class AsyncEmbeddings(AsyncAPIResource, UseCaseResource):  # type: ignore[misc]
    """
    Resource to interact with embeddings.
    """

    def __init__(self, client: AsyncAdaptive) -> None:
        AsyncAPIResource.__init__(self, client)
        UseCaseResource.__init__(self, client)

    async def create(
        self,
        input: str,
        model_key: str | None = None,
        encoding_format: Literal["Float", "Base64"] = "Float",
        use_case: str | None = None,
        user: str | UUID | None = None,
    ) -> rest_types.EmbeddingsResponseList:
        """
        Creates embeddings inference request.

        Args:
            input: Input text to embed.
            model: Target model key for inference. If `None`, the requests will be routed to the use case's default model.
                Request will error if default model is not an embedding model.
            encoding_format: Encoding format of response.
            user: ID of user making the requests. If not `None`, will be logged as metadata for the request.
        """
        encoding_format_enum = rest_types.EmbeddingsEncodingFormat(encoding_format)
        emb_input = rest_types.GenerateEmbeddingsInput(
            input=input,
            model=get_full_model_path(self.use_case_key(use_case), model_key),
            encoding_format=encoding_format_enum,
            dimensions=None,
            user=convert_optional_UUID(user),
        )
        r = await self._rest_client.post(
            ROUTE, data=emb_input.model_dump_json(exclude_none=True), headers={"Content-Type": "application/json"}  # type: ignore
        )
        rest_error_handler(r)
        return rest_types.EmbeddingsResponseList.model_validate(r.json())



from __future__ import annotations
from typing import TYPE_CHECKING, List

from adaptive_sdk.graphql_client import ListPermissions

from .base_resource import SyncAPIResource, AsyncAPIResource, UseCaseResource

if TYPE_CHECKING:
    from adaptive_sdk.client import Adaptive, AsyncAdaptive


class Permissions(SyncAPIResource, UseCaseResource):  # type: ignore[misc]
    """
    Resource to list permissions.
    """

    def __init__(self, client: Adaptive) -> None:
        SyncAPIResource.__init__(self, client)
        UseCaseResource.__init__(self, client)

    def list(self) -> List[str]:
        return self._gql_client.list_permissions().permissions


class AsyncPermissions(AsyncAPIResource, UseCaseResource):  # type: ignore[misc]
    """
    Resource to list permissions.
    """

    def __init__(self, client: AsyncAdaptive) -> None:
        AsyncAPIResource.__init__(self, client)
        UseCaseResource.__init__(self, client)

    async def list(self) -> List[str]:
        return (await self._gql_client.list_permissions()).permissions



from pydantic import Field
from .base_model import BaseModel
from .fragments import JudgeData

class CreateJudge(BaseModel):
    """@public"""
    create_judge: 'CreateJudgeCreateJudge' = Field(alias='createJudge')

class CreateJudgeCreateJudge(JudgeData):
    """@public"""
    pass
CreateJudge.model_rebuild()


from typing import List, Optional
from pydantic import Field
from .base_model import BaseModel
from .fragments import JobData

class ListJobs(BaseModel):
    """@public"""
    jobs: 'ListJobsJobs'

class ListJobsJobs(BaseModel):
    """@public"""
    total_count: int = Field(alias='totalCount')
    page_info: 'ListJobsJobsPageInfo' = Field(alias='pageInfo')
    nodes: List['ListJobsJobsNodes']

class ListJobsJobsPageInfo(BaseModel):
    """@public"""
    has_next_page: bool = Field(alias='hasNextPage')
    has_previous_page: bool = Field(alias='hasPreviousPage')
    start_cursor: Optional[str] = Field(alias='startCursor')
    end_cursor: Optional[str] = Field(alias='endCursor')

class ListJobsJobsNodes(JobData):
    """@public"""
    pass
ListJobs.model_rebuild()
ListJobsJobs.model_rebuild()


from abc import ABC, abstractmethod
import asyncio
from adaptive_sdk.external.requests_journal import RequestsJournal
from adaptive_sdk.external.websocket_utils import send_large_text
from fastapi import FastAPI, APIRouter, WebSocket
from fastapi.responses import JSONResponse
from typing import Callable, Any, Generic, Type, TypeVar
from dataclasses import dataclass
import threading
from pydantic import Field
import uvicorn
from loguru import logger

from adaptive_sdk.external.reward_types import (
    ValidatedRequest,
    Response,
    ServerInfo,
    ValidatedBatchedRequest,
    BatchedResponse,
    BaseModel,
)
from adaptive_sdk.external.constants import (
    SCORE_PATH,
    BATCH_SCORE_PATH,
    INFO_PATH,
    METADATA_SCHEMA_PATH,
)


class EmptyMetadata(BaseModel):
    pass


@dataclass
class Route:
    path: str
    endpoint: Callable[..., Any]
    methods: list[str]


META = TypeVar("META", bound=BaseModel)


class RewardServer(ABC, Generic[META]):
    def __init__(
        self,
        port: int,
        metadata_cls: Type[META],
        blocking: bool = True,
        verbose: bool = False,
        request_timeout_s: int = 3_600,
        requests_journal: RequestsJournal | None = None,
    ):
        self.requests_journal = requests_journal
        self.request_timeout_s = request_timeout_s
        self.metadata_cls = metadata_cls
        # need to update the functions with the correct metadata type, otherwise the bound BaseModel is used
        # (see https://github.com/fastapi/fastapi/issues/5874)
        self._score.__annotations__["request"] = ValidatedRequest[metadata_cls]  # type: ignore[valid-type]
        self.batch_score.__annotations__["requests"] = ValidatedBatchedRequest[metadata_cls]  # type: ignore[valid-type]

        self.verbose = verbose
        self.__post_init__()
        self._setup_server(port, blocking)
        self.cleanup()

    def __post_init__(self):
        pass

    def cleanup(self):
        pass

    async def _score(self, request: ValidatedRequest[META]) -> Response:
        if self.verbose:
            request_str = request.model_dump_json(indent=2)
            print(f"Received request: {request_str}")
            res = await self.score(request)
            print(f"Finished request : {request_str} with response: {res.model_dump_json(indent=2)}")
            return res
        else:
            return await self.score(request)

    @abstractmethod
    async def score(self, request: ValidatedRequest[META]) -> Response: ...

    async def batch_score(self, requests: ValidatedBatchedRequest[META]) -> BatchedResponse:
        tasks = []
        for request in requests.requests:
            tasks.append(self._score(request))
        responses = await asyncio.gather(*tasks)
        return BatchedResponse(responses=responses)

    def get_medata_schema(self) -> JSONResponse:
        json_schema = self.metadata_cls.model_json_schema()
        return JSONResponse(json_schema)

    @abstractmethod
    async def info(self) -> ServerInfo: ...

    def _setup_server(self, port: int, blocking: bool, log_request: bool = False):
        class ThreadServer(uvicorn.Server):
            """ "Easy to kill uvicorn server"""

            def run_in_thread(self):
                self.thread = threading.Thread(target=self.run, daemon=True)
                self.thread.start()

            def stop(self):
                self.should_exit = True
                self.thread.join()

        def get_routes() -> list[Route]:
            routes: list[Route] = []
            routes.append(Route(SCORE_PATH, self._score, methods=["POST"]))
            routes.append(Route(BATCH_SCORE_PATH, self.batch_score, methods=["POST"]))
            routes.append(Route(INFO_PATH, self.info, methods=["GET"]))
            routes.append(Route(METADATA_SCHEMA_PATH, self.get_medata_schema, methods=["GET"]))
            return routes

        async def score_task(websocket: WebSocket, msg):
            req: ValidatedRequest = ValidatedRequest[self.metadata_cls].model_validate(msg) # type: ignore
            req_id: int = req.id # type: ignore
            response = await self._score(req)
            response.id = req_id
            await send_large_text(websocket, response.model_dump_json(), id=req_id)
            # await websocket.send_text(response.model_dump_json())

        async def websocket_endpoint(websocket: WebSocket):
            await websocket.accept()
            tasks = set()
            while True:
                try:
                    json = await websocket.receive_json()
                    task = asyncio.create_task(score_task(websocket, json))
                    tasks.add(task)
                    task.add_done_callback(tasks.discard)

                except Exception as e:
                    logger.error(f"{e}")
                    break

        router = APIRouter()
        routes = get_routes()

        for route in routes:
            router.add_api_route(route.path, route.endpoint, methods=route.methods)

        app = FastAPI()
        app.include_router(router)

        app.add_api_websocket_route("/ws", websocket_endpoint)

        if self.requests_journal is not None:
            self.requests_journal.add_journalling(app=app)

        config = uvicorn.Config(
            app,
            host="0.0.0.0",
            port=port,
            log_level="info",
            # We do not use workers here. Depending on how the app is launched
            # they are not used. Which makes it misleading.
            ws_ping_timeout=None,
        )
        self.server = ThreadServer(config=config)

        if blocking:
            self.server.run()
        else:
            self.server.run_in_thread()


class MyMetadata(BaseModel):
    scary_letter: str = Field(min_length=1, max_length=1)


class MyRewardServer(RewardServer[MyMetadata]):
    def __init__(self, port=8000, blocking=True):
        super().__init__(port, MyMetadata, blocking, requests_journal=RequestsJournal())

    async def score(self, request: ValidatedRequest[MyMetadata]) -> Response:

        last_completion = request.turns[-1].content
        num_scary_letters = last_completion.count(request.metadata.scary_letter)
        return Response(
            reward=0.0 if request.metadata.scary_letter in last_completion else 1.0,
            metadata={
                "feedback": (
                    "There were no scary letters!"
                    if num_scary_letters == 0
                    else f"There were {num_scary_letters} scary letters!"
                )
            },
        )

    async def info(self) -> ServerInfo:
        return ServerInfo(version="1.0", name="My sevice", description="This is a nice description")


class MyNoMetadataRewardServer(RewardServer[EmptyMetadata]):
    def __init__(self, port=8000, blocking=True):
        super().__init__(port, EmptyMetadata, blocking, requests_journal=RequestsJournal())

    async def score(self, request: ValidatedRequest[EmptyMetadata]) -> Response:

        last_completion = request.turns[-1].content
        return Response(reward=len(last_completion), metadata={})

    async def info(self) -> ServerInfo:
        return ServerInfo(version="1.0", name="My sevice", description="This is a nice description")


if __name__ == "__main__":
    server = MyRewardServer(port=50056)



from pydantic import Field
from .base_model import BaseModel
from .fragments import AbCampaignCreateData

class CreateAbCampaign(BaseModel):
    """@public"""
    create_ab_campaign: 'CreateAbCampaignCreateAbCampaign' = Field(alias='createAbCampaign')

class CreateAbCampaignCreateAbCampaign(AbCampaignCreateData):
    """@public"""
    pass
CreateAbCampaign.model_rebuild()


from pydantic import Field
from .base_model import BaseModel
from .fragments import JobData

class CancelJob(BaseModel):
    """@public"""
    cancel_job: 'CancelJobCancelJob' = Field(alias='cancelJob')

class CancelJobCancelJob(JobData):
    """@public"""
    pass
CancelJob.model_rebuild()


from typing import List, Optional
from pydantic import Field
from .base_model import BaseModel
from .enums import CompletionGroupBy
from .fragments import CompletionData, MetricData

class ListGroupedInteractions(BaseModel):
    """@public"""
    completions_grouped: 'ListGroupedInteractionsCompletionsGrouped' = Field(alias='completionsGrouped')

class ListGroupedInteractionsCompletionsGrouped(BaseModel):
    """@public"""
    total_count: int = Field(alias='totalCount')
    group_by: CompletionGroupBy = Field(alias='groupBy')
    page_info: 'ListGroupedInteractionsCompletionsGroupedPageInfo' = Field(alias='pageInfo')
    nodes: List['ListGroupedInteractionsCompletionsGroupedNodes']

class ListGroupedInteractionsCompletionsGroupedPageInfo(BaseModel):
    """@public"""
    has_next_page: bool = Field(alias='hasNextPage')
    end_cursor: Optional[str] = Field(alias='endCursor')

class ListGroupedInteractionsCompletionsGroupedNodes(BaseModel):
    """@public"""
    key: Optional[str]
    count: int
    direct_feedbacks_stats: List['ListGroupedInteractionsCompletionsGroupedNodesDirectFeedbacksStats'] = Field(alias='directFeedbacksStats')
    completions: 'ListGroupedInteractionsCompletionsGroupedNodesCompletions'

class ListGroupedInteractionsCompletionsGroupedNodesDirectFeedbacksStats(BaseModel):
    """@public"""
    metric: 'ListGroupedInteractionsCompletionsGroupedNodesDirectFeedbacksStatsMetric'
    feedbacks: int
    average: Optional[float]
    max: Optional[float]
    min: Optional[float]
    stddev: Optional[float]
    sum: Optional[float]

class ListGroupedInteractionsCompletionsGroupedNodesDirectFeedbacksStatsMetric(MetricData):
    """@public"""
    pass

class ListGroupedInteractionsCompletionsGroupedNodesCompletions(BaseModel):
    """@public"""
    nodes: List['ListGroupedInteractionsCompletionsGroupedNodesCompletionsNodes']

class ListGroupedInteractionsCompletionsGroupedNodesCompletionsNodes(CompletionData):
    """@public"""
    pass
ListGroupedInteractions.model_rebuild()
ListGroupedInteractionsCompletionsGrouped.model_rebuild()
ListGroupedInteractionsCompletionsGroupedNodes.model_rebuild()
ListGroupedInteractionsCompletionsGroupedNodesDirectFeedbacksStats.model_rebuild()
ListGroupedInteractionsCompletionsGroupedNodesCompletions.model_rebuild()


from pydantic import Field
from .base_model import BaseModel
from .fragments import UserData

class CreateUser(BaseModel):
    """@public"""
    create_user: 'CreateUserCreateUser' = Field(alias='createUser')

class CreateUserCreateUser(UserData):
    """@public"""
    pass
CreateUser.model_rebuild()


from typing import Literal, Union
from pydantic import Field
from .base_model import BaseModel

class TestRemoteEnv(BaseModel):
    """@public"""
    test_remote_env: Union['TestRemoteEnvTestRemoteEnvRemoteEnvTestOffline', 'TestRemoteEnvTestRemoteEnvRemoteEnvTestOnline'] = Field(alias='testRemoteEnv', discriminator='typename__')

class TestRemoteEnvTestRemoteEnvRemoteEnvTestOffline(BaseModel):
    """@public"""
    typename__: Literal['RemoteEnvTestOffline'] = Field(alias='__typename')
    error: str

class TestRemoteEnvTestRemoteEnvRemoteEnvTestOnline(BaseModel):
    """@public"""
    typename__: Literal['RemoteEnvTestOnline'] = Field(alias='__typename')
    name: str
    version: str
    description: str
TestRemoteEnv.model_rebuild()


from pydantic import Field
from .base_model import BaseModel
from .fragments import MetricWithContextData

class LinkMetric(BaseModel):
    """@public"""
    link_metric: 'LinkMetricLinkMetric' = Field(alias='linkMetric')

class LinkMetricLinkMetric(MetricWithContextData):
    """@public"""
    pass
LinkMetric.model_rebuild()


from typing import Optional
from .base_model import BaseModel
from .fragments import DatasetData

class DescribeDataset(BaseModel):
    """@public"""
    dataset: Optional['DescribeDatasetDataset']

class DescribeDatasetDataset(DatasetData):
    """@public"""
    pass
DescribeDataset.model_rebuild()


from typing import List
from .base_model import BaseModel
from .fragments import JudgeData

class ListJudges(BaseModel):
    """@public"""
    judges: List['ListJudgesJudges']

class ListJudgesJudges(JudgeData):
    """@public"""
    pass
ListJudges.model_rebuild()


from .graphql_client import BaseModel
from pprint import pformat


def pretty_pydantic_print(self):
    return pformat(self.model_dump(), indent=2)


BaseModel.__str__ = pretty_pydantic_print  # type: ignore[method-assign]



from pydantic import Field
from .base_model import BaseModel
from .fragments import UserData

class DeleteUser(BaseModel):
    """@public"""
    delete_user: 'DeleteUserDeleteUser' = Field(alias='deleteUser')

class DeleteUserDeleteUser(UserData):
    """@public"""
    pass
DeleteUser.model_rebuild()


from adaptive_sdk.external.websocket_utils import ResponseAccumulator
import httpx
from typing import Any
from typing_extensions import Self
from httpx import Limits
from jsonschema import validate, ValidationError as JsonSchemaValidationError
from websockets.asyncio.client import connect, ClientConnection
import asyncio
import json
from loguru import logger

from adaptive_sdk.external.reward_types import (
    BatchedMetadataValidationResponse,
    MetadataValidationResponse,
    Request,
    Response,
    ServerInfo,
    BatchedRequest,
    BatchedResponse,
    Turn,
)
from adaptive_sdk.external.constants import (
    METADATA_SCHEMA_PATH,
    SCORE_PATH,
    BATCH_SCORE_PATH,
    INFO_PATH,
)


async def read_task(client: ClientConnection, responses: dict[int, Response | asyncio.Event]):
    response_accumulators: dict[int, ResponseAccumulator] = {}
    while True:
        try:
            msg = await client.recv()
            obj = json.loads(msg)
            id = obj["id"]
            if total_num_chunks := obj.get("total_num_chunks"):
                assert id not in response_accumulators
                response_accumulators[id] = ResponseAccumulator(total_num_chunks)
            else:
                assert (acc := response_accumulators.get(id))
                acc.add_chunk(obj["chunk"])
                if acc.is_complete():
                    del response_accumulators[id]
                    response = Response.model_validate_json(acc.get_full_data())  # type: ignore
                    # sanity check
                    assert response.id == id
                    event: asyncio.Event = responses.pop(id)  # type: ignore
                    responses[id] = response
                    event.set()
        except Exception as e:
            logger.error(f"{e}")
            break


class RewardClient:
    def __init__(self, base_url: str, max_connections: int = 32, timeout: float | None = None):
        headers = dict()  # type: ignore[var-annotated]
        self.base_url = base_url.removeprefix("http://").removeprefix("https://")
        self._client = httpx.AsyncClient(
            headers=dict(),
            base_url=f"http://{self.base_url}",
            timeout=timeout,
            limits=Limits(max_connections=max_connections),
        )
        self._metadata_json_schema: None | dict[str, Any] = None
        self.use_websocket = False
        self.max_connections = max_connections

    async def setup(self):
        await self.connect_websocket()

    async def connect_websocket(self)->Self:
        self.use_websocket = True
        self.ws_client: ClientConnection = await connect(f"ws://{self.base_url}/ws", ping_timeout=None)
        self.ws_responses: dict[int, Response | asyncio.Event] = dict()
        logger.info("Spawning_read_task")
        self.read_task = asyncio.create_task(read_task(self.ws_client, self.ws_responses))
        self.request_id = 0
        # no need to blast 2x more than the amount of workers
        self.semaphore = asyncio.Semaphore(self.max_connections)
        return self

    async def drop_websocket(self):
        assert self.use_websocket
        logger.info("Cancelling_read_task")
        self.read_task.cancel()
        await self.ws_client.close()

    async def _post(self, path: str, data: dict) -> httpx.Response:
        response = await self._client.post(path, json=data)
        response.raise_for_status()
        return response

    async def _ws_post(self, req: Request) -> Response:
        async with self.semaphore:
            request_id = self.request_id
            self.request_id += 1
            req.id = request_id
            event = asyncio.Event()
            self.ws_responses[request_id] = event
            await self.ws_client.send(req.model_dump_json())
            await event.wait()
            response: Response = self.ws_responses.pop(request_id)  # type: ignore
            assert response.id == request_id
            return response

    async def score(self, req: Request) -> Response:
        if not self.use_websocket:
            response = await self._post(SCORE_PATH, req.model_dump())
            return Response(**response.json())
        else:
            return await self._ws_post(req)

    async def batch_score(self, requests: list[Request]) -> list[Response]:
        response = await self._post(BATCH_SCORE_PATH, BatchedRequest(requests=requests).model_dump())
        batched_response = BatchedResponse(**response.json())
        return batched_response.responses

    async def validate_metadata(self, metadata: dict[Any, Any]):
        if self._metadata_json_schema is None:
            response = await self._client.get(METADATA_SCHEMA_PATH)
            self._metadata_json_schema = response.json()

        try:
            validate(instance=metadata, schema=self._metadata_json_schema)  # type: ignore
            return MetadataValidationResponse(is_valid=True)
        except JsonSchemaValidationError as e:
            return MetadataValidationResponse(is_valid=False, error_message=str(e))

    async def batch_validate_metadata(
        self, list_of_metadata: list[dict[Any, Any]]
    ) -> BatchedMetadataValidationResponse:
        return BatchedMetadataValidationResponse(
            responses=[await self.validate_metadata(metadata) for metadata in list_of_metadata]
        )

    async def info(self) -> ServerInfo:
        response = await self._client.get(INFO_PATH)
        response.raise_for_status()
        return ServerInfo(**response.json())

    def blocking_info(self) -> ServerInfo:
        return ServerInfo(**httpx.get(self._client.base_url.join(INFO_PATH), timeout=self._client.timeout).json())

    def blocking_batch_score(self, requests: list[Request]) -> list[Response]:
        request = BatchedRequest(requests=requests)
        response = BatchedResponse(
            **httpx.post(
                self._client.base_url.join(BATCH_SCORE_PATH),
                json=request.model_dump(),
                timeout=self._client.timeout,
            ).json()
        )
        return response.responses


async def main():
    # client = RewardClient("0.0.0.0:50056")
    client = await RewardClient("0.0.0.0:50056").connect_websocket()
    tasks = []
    for _ in range(1024):
        task = client.score(Request(turns=[Turn(role="assistant", content="hello")]))
        tasks.append(task)
    await asyncio.gather(*tasks)


if __name__ == "__main__":
    asyncio.run(main())



from typing import Optional
from .base_model import BaseModel
from .fragments import ModelData

class DescribeModel(BaseModel):
    """@public"""
    model: Optional['DescribeModelModel']

class DescribeModelModel(ModelData):
    """@public"""
    backbone: Optional['DescribeModelModelBackbone']

class DescribeModelModelBackbone(ModelData):
    """@public"""
    pass
DescribeModel.model_rebuild()
DescribeModelModel.model_rebuild()


from typing import Optional
from pydantic import Field
from .base_model import BaseModel
from .fragments import ModelData

class AddModel(BaseModel):
    """@public"""
    add_model: 'AddModelAddModel' = Field(alias='addModel')

class AddModelAddModel(ModelData):
    """@public"""
    backbone: Optional['AddModelAddModelBackbone']

class AddModelAddModelBackbone(ModelData):
    """@public"""
    pass
AddModel.model_rebuild()
AddModelAddModel.model_rebuild()